@@ -1,4 +1,4 @@
-/* Copyright 2001, 2002 Manuel Arriaga
+/* Copyright 2001, 2002, 2003 Manuel Arriaga
  *
  *
  * This program is free software; you can redistribute it and/or modify
@@ -31,12 +31,16 @@
 #include <unistd.h>
 #include <pwd.h>
 #include <sys/stat.h>
+#include <regex.h>
 
 #include "trash.h"
 
+/* This file contains a series of helper functions, which the wrappers use. Their
+ * definition is preceded by a short explanation of what they do. */
+
 /* --------------------------------------------- */
 
-/* There isn't any need to export these functions: */
+/* There isn't any need to export these functions (they are mere "helper helper functions" :-) ): */
 
 static char** read_config_from_file(const char *conf_file, int number_of_keys, config *cfg, ...);
 
@@ -46,14 +50,24 @@ static int reformulate_new_path(char **new_path, char **first_null);
 
 static int move(const char *old_path, const char *new_path, config *cfg);
 
+static int is_an_exception(const char *path, const char *exceptions);
+
+static int is_empty_file(const char *path, config *cfg);
+
+static int matches_re(const char *path, const char *regexp);
+
 /* Definition of helper functions: */
 
 /* ------------------------------------------------------------------------ */
 
 /* This function returns 1 if the given path points to a file or directory under one
  * of the directories listed in the semi-colon separated dir_list, 0 otherwise. We don't
- * use strtok() because doing so would involve creating a copy of dir_list in memory,
- * something which would expose us to nasty exception handling problems.
+ * use strtok_r() because doing so would involve creating both a copy of dir_list and an
+ * empty buffer of the same size in memory, and, quite frankly, checking for an error code
+ * after each call to found_under_dir() in libtrash would make the code very ugly. (That's what
+ * would be required if this function depended on malloc()ing memory, an operation which might
+ * fail.) It is meant to be a "primitive" which can be used without having to worry about internal
+ * failure.
  */
 
 int found_under_dir(const char *abs_path, const char *dir_list)
@@ -74,17 +88,28 @@ int found_under_dir(const char *abs_path, const char *dir_list)
 	semi_colon = strchr(beg_name, ';');
 
 	if (semi_colon)
-	  end_name = semi_colon;
-	else /* if there aren't any more semi-colons, point end_name to the '\0' at the end of dir_list: */
-	  end_name = dir_list + strlen(dir_list);
+	  {
+	     if (semi_colon > dir_list && /* we don't want to accidentally access forbidden mem areas */
+		 *(semi_colon - 1) == '/') /* dir name has a trailing slash, ignore it */
+	       end_name = semi_colon - 1;
+	     else /* the first char after the dir name is really the semi-colon */
+	       end_name = semi_colon;
+	  }
+	else /* if there aren't any more semi-colons, point it to the '\0' at the end of the string or the last '/': */
+	  {
+	     if (dir_list[strlen(dir_list) - 1] == '/')
+	       end_name = dir_list + strlen(dir_list) - 1; /* '/' is the char after the last dir name */
+	     else
+	       end_name = dir_list + strlen(dir_list); /* '\0' is the last char after the last dir name */
+	  }
 
 	if (!strncmp(abs_path, beg_name, end_name - beg_name) && *(abs_path + (end_name - beg_name)) == '/')
 	  return 1;
 
 	if (semi_colon)
-	  beg_name = end_name + 1;
+	  beg_name = semi_colon + 1;
 	else /* if there aren't any more semi-colons, point beg_name to the '\0' at the end of dir_list: */
-	  beg_name = end_name;
+	  beg_name = dir_list + strlen(dir_list);
      }
 
    /* If we get to this point, return signalling that abs_path isn't located under any of the dirs in dir_list: */
@@ -95,6 +120,49 @@ int found_under_dir(const char *abs_path, const char *dir_list)
 
 /* ---------------------------------------------------------------------- */
 
+/* This function is very similar to found_under_dir: it simply checks whether path starts with one
+ * of the file paths listed in exceptions. If it does, this function returns 1; if it doesn't, it
+ * returns 0. */
+
+static int is_an_exception(const char *path, const char *exceptions)
+{
+
+   const char *beg_name = NULL, *end_name = NULL;
+   const char *semi_colon = NULL;
+
+   beg_name = exceptions;
+
+   /* Special case: exceptions is NULL, meaning that the user didn't provide us with such a list of directories.
+    * Just return 0: */
+
+   if (exceptions == NULL)
+     return 0;
+
+   while (*beg_name != '\0')
+     {
+	semi_colon = strchr(beg_name, ';');
+
+	if (semi_colon)
+	  end_name = semi_colon;
+	else /* if there aren't any more semi-colons, point it to the '\0' at the end of the string: */
+	  end_name = exceptions + strlen(exceptions); /* '\0' is the last char after the last dir name */
+
+	if (!strncmp(path, beg_name, end_name - beg_name)) /* path is equal to (or at least begins with) a path found */
+	  return 1;                                        /* in EXCEPTIONS */
+
+	if (semi_colon)
+	  beg_name = semi_colon + 1;
+	else /* if there aren't any more semi-colons, point beg_name to the '\0' at the end of cfg->exceptions: */
+	  beg_name = exceptions + strlen(exceptions);
+     }
+
+   /* If we get to this point, return signalling that path isn't one of the exceptions: */
+
+   return 0;
+}
+
+/* --------------------------------------------- */
+
 /* This function tests the existence and permissions of the dir's pathname; if it exists and has
  * the right permissions, it returns 1. If it exists but has the wrong
  * permissions and we can't correct them, or if pathname points to a file,
@@ -636,30 +704,16 @@ int reformulate_new_path(char **new_path, char **first_null)
  * begins with a dot) or resides under a hidden directory (i.e., under a directory whose
  * name begins with a dot). It returns 0 otherwise.
  *
- * How we do it: we search for all the occurrences of the substring "/." in the path we were
- * given, and then see if any of them is followed by a character other than a slash or a (second)
- * dot. If this is the case, that "/." marks the beginning of a hidden directory's or file's
- * name, and we return 1. If no single "/." is found, or all found ones are followed by a slash
- * or a dot, we return 0.
+ * Because absolute_path has already been "canonicalized", there's no need to worry about symlinks such as ".." and ".".
+ *
  */
 
 int hidden_file(const char *absolute_path)
 {
-
-   const char *ptr = absolute_path;
-
-   while ((ptr = strstr(absolute_path, "/.")))
-     {
-	/* By adding 2 to ptr we get a pointer to the char which determines whether this is
-	 * a hidden file/dir's name or just a reference to the "/.." or "/./": */
-
-	if (*(ptr + 2) == '/' || *(ptr + 2) == '.')
-	  continue;
-	else
-	  return 1;
-     }
-
-   return 0;
+   if (strstr(absolute_path, "/."))
+     return 1;
+   else
+     return 0;
 }
 
 /* ---------------------------------------------------------------------------- */
@@ -678,7 +732,7 @@ int hidden_file(const char *absolute_path)
  * This makes it useless in the unlink() and rename() wrappers, because those functions
  * act upon the symlinks themselves, not the files they point at. But they still need to
  * resolve any other symlinks in the path, so we can't just use the absolute "unresolved"
- path.
+ * path.
  */
 
 char* build_absolute_path(const char *path)
@@ -1217,7 +1271,7 @@ static char * readline(FILE *stream, int *errors)
 	return NULL;
      }
 
-   if (strlen(line) == 0)
+   if (strlen(line) == 0 && feof(stream))
      {
 	free(line);
 	return NULL;
@@ -1236,9 +1290,8 @@ static char * readline(FILE *stream, int *errors)
 
 /* These macros are used by the function get_config_from_file(): */
 
-#define NUMBER_OF_CONFIG_OPTIONS 19 /* number of options listed below in the call to read_config_from_file(). This
-				     value is also read by the (Perl) script which composes the system-wide configuration
-				     file. */
+#define NUMBER_OF_CONFIG_OPTIONS 22 /* number of options listed below in the call to read_config_from_file().*/
+
 /* ---------------------------- */
 
 #define SET_INTEGER_FREE_MEMORY(var,str,str1,str2,int1,int2,intdef)                 \
@@ -1276,16 +1329,19 @@ void get_config_from_file(config *cfg)
 					 "GLOBAL_PROTECTION",
 					 "TRASH_SYSTEM_ROOT",
 					 "TEMPORARY_DIRS",
+					 "USER_TEMPORARY_DIRS",
 					 "UNREMOVABLE_DIRS",
 					 "IGNORE_EXTENSIONS",
-					 "UNLINK_OFF",
-					 "RENAME_OFF",
-					 "FOPEN_OFF",
-					 "FREOPEN_OFF",
-					 "OPEN_OFF",
+					 "INTERCEPT_UNLINK",
+					 "INTERCEPT_RENAME",
+					 "INTERCEPT_FOPEN",
+					 "INTERCEPT_FREOPEN",
+					 "INTERCEPT_OPEN",
 					 "LIBTRASH_CONFIG_FILE_UNREMOVABLE",
 					 "REMOVABLE_MEDIA_MOUNT_POINTS",
-					 "IGNORE_EDITOR_TEMPORARY");
+					 "IGNORE_EDITOR_TEMPORARY",
+					 "EXCEPTIONS",
+					 "IGNORE_RE");
 
    /* Did read_config_from_file() fail? If it did, we quit and leave the compile-time defaults unchanged: */
 
@@ -1362,44 +1418,56 @@ void get_config_from_file(config *cfg)
      cfg->temporary_dirs = config_values[8];
 
    if (config_values[9])
-     cfg->unremovable_dirs = config_values[9];
+     cfg->user_temporary_dirs = config_values[9];
 
    if (config_values[10])
-     cfg->ignore_extensions = config_values[10];
+     cfg->unremovable_dirs = config_values[10];
+
+   if (config_values[11])
+     cfg->ignore_extensions = config_values[11];
 
    /* (All of the memory pointed to by these three pointers will be free()d in _fini().)  */
 
    /* Should we (dis)enable any function specifically? */
 
-   SET_INTEGER_FREE_MEMORY(cfg->unlink_off, config_values[11], "YES", "NO",
-			   YES, NO, UNLINK_OFF);
+   SET_INTEGER_FREE_MEMORY(cfg->intercept_unlink, config_values[12], "YES", "NO",
+			   YES, NO, INTERCEPT_UNLINK);
 
-   SET_INTEGER_FREE_MEMORY(cfg->rename_off, config_values[12], "YES", "NO",
-			   YES, NO, RENAME_OFF);
+   SET_INTEGER_FREE_MEMORY(cfg->intercept_rename, config_values[13], "YES", "NO",
+			   YES, NO, INTERCEPT_RENAME);
 
-   SET_INTEGER_FREE_MEMORY(cfg->fopen_off, config_values[13], "YES", "NO",
-			   YES, NO, FOPEN_OFF);
+   SET_INTEGER_FREE_MEMORY(cfg->intercept_fopen, config_values[14], "YES", "NO",
+			   YES, NO, INTERCEPT_FOPEN);
 
-   SET_INTEGER_FREE_MEMORY(cfg->freopen_off, config_values[14], "YES", "NO",
-			   YES, NO, FREOPEN_OFF);
+   SET_INTEGER_FREE_MEMORY(cfg->intercept_freopen, config_values[15], "YES", "NO",
+			   YES, NO, INTERCEPT_FREOPEN);
 
-   SET_INTEGER_FREE_MEMORY(cfg->open_off, config_values[15], "YES", "NO",
-			   YES, NO, OPEN_OFF);
+   SET_INTEGER_FREE_MEMORY(cfg->intercept_open, config_values[16], "YES", "NO",
+			   YES, NO, INTERCEPT_OPEN);
 
    /* Should we allow the destruction of the user's libtrash configuration file? */
 
-   SET_INTEGER_FREE_MEMORY(cfg->libtrash_config_file_unremovable, config_values[16], "YES", "NO",
+   SET_INTEGER_FREE_MEMORY(cfg->libtrash_config_file_unremovable, config_values[17], "YES", "NO",
 			   YES, NO, LIBTRASH_CONFIG_FILE_UNREMOVABLE);
 
    /* Are there any directories under which files should be really destroyed because they are used as mount-points
     * for removable media? (equivalent to temporary_dirs) */
 
-   if (config_values[17])
-     cfg->removable_media_mount_points = config_values[17];
+   if (config_values[18])
+     cfg->removable_media_mount_points = config_values[18];
 
-   SET_INTEGER_FREE_MEMORY(cfg->ignore_editor_temporary, config_values[18], "YES", "NO",
+   SET_INTEGER_FREE_MEMORY(cfg->ignore_editor_temporary, config_values[19], "YES", "NO",
 			   YES, NO, IGNORE_EDITOR_TEMPORARY);
 
+   /* Are there any files which would typically be covered by libtrash which the user decided to list
+    * as "exceptions"? */
+
+   if (config_values[20])
+     cfg->exceptions = config_values[20];
+
+   if (config_values[21])
+     cfg->ignore_re = config_values[21];
+
    /* Done. */
 
    /* We no longer need the config_values array, since we already copied/used
@@ -1414,7 +1482,7 @@ void get_config_from_file(config *cfg)
 
 /* ----------------------------------------------------------------------------------- */
 
-/* decide_action() takes an absolute canonical path as its sole argument and indicates
+/* decide_action() takes an absolute canonical path and the current config settings as its arguments and indicates
  to the caller how this file should be handled. It returns one of three possible values:
  *
  - BE_REMOVED: according to the user's preferences, this file can be removed without having
@@ -1431,25 +1499,29 @@ int decide_action(const char *absolute_path, config *cfg)
    const char *tmp = NULL;
 
    /* Tell the caller to handle the files already under the user's trash can according to the
-    value of cfg->protect_trash: */
+    value of cfg->protect_trash, also taking into consideration whether (or not) the trash can
+    is currently listed in UNCOVER_DIRS: */
 
    if (found_under_dir(absolute_path, cfg->absolute_trash_can))
      {
-	if (cfg->protect_trash == NO)
+	if (cfg->protect_trash == NO ||
+	    found_under_dir(absolute_path, cfg->uncovered_dirs)) /* user temporarily disabled PROTECT_TRASH via UNCOVER_DIRS */
 	  return BE_REMOVED;
-	else /* if cfg->protect_trash == YES */
+	else /* if cfg->protect_trash == YES && !found_under_dir(absolute_path, cfg->uncover_dirs) */
 	  return BE_LEFT_UNTOUCHED;
      }
 
    /* Tell the caller to return an error code and don't even touch these files: */
 
-   if (found_under_dir(absolute_path, cfg->unremovable_dirs) || /* (a) */
-       (cfg->libtrash_config_file_unremovable                && /* (b) */
-	(found_under_dir(absolute_path, cfg->home) && !strcmp(absolute_path + strlen(cfg->home) + 1, PERSONAL_CONF_FILE)) ) )
-     return BE_LEFT_UNTOUCHED;                             /* (c) */
+   if ( (found_under_dir(absolute_path, cfg->unremovable_dirs) &&
+	 !found_under_dir(absolute_path, cfg->uncovered_dirs)  &&
+	 !is_an_exception(absolute_path, cfg->exceptions) ) || /* (a) */
+	(cfg->libtrash_config_file_unremovable                && /* (b) */
+	 (found_under_dir(absolute_path, cfg->home) && !strcmp(absolute_path + strlen(cfg->home) + 1, PERSONAL_CONF_FILE)) ) ) /* (c) */
+     return BE_LEFT_UNTOUCHED;
 
    /* Notes:
-    * (a) - file lies in one of the cfg->unremovable_dirs or
+    * (a) - file lies in one of the cfg->unremovable_dirs, that dir hasn't been "uncovered" and this file isn't an "exception"; or
     * (b) - we have instructions to protect the user's libtrash configuration file and
     * (c) - we make sure that this file is in the user's home directory and then compare PERSONAL_CONF_FILE with the portion
     * of absolute_path which _follows_ the name of the home directory and the slash which separates it from the file name.
@@ -1464,16 +1536,24 @@ int decide_action(const char *absolute_path, config *cfg)
 
 	(cfg->ignore_editor_temporary &&                                             /* is a temporary file used by a text */
 	 ((tmp = strrchr(absolute_path, '/')) ? *(tmp + 1) == '#' :                  /* editor */
-	  absolute_path[0] == '#') ) ||
+	  absolute_path[0] == '#') )                                            ||
 
-	found_under_dir(absolute_path, cfg->temporary_dirs) ||                           /* is a (normal) temporary file; */
+	found_under_dir(absolute_path, cfg->temporary_dirs)                     ||      /* is a (normal) temporary file; */
 
-	(!found_under_dir(absolute_path, cfg->home) && !cfg->global_protection) ||           /* is outside of the user's dir and
-											      the user doesn't want us to protect
-											      these files. */
-	ends_in_ignored_extension(absolute_path, cfg)  ||                               /* filename ends in an extension
+	found_under_dir(absolute_path, cfg->user_temporary_dirs)                ||      /* is a temporary file in a dir under $HOME; */
+
+	(!found_under_dir(absolute_path, cfg->home) && !cfg->global_protection) ||      /* is outside of the user's dir and
+											 the user doesn't want us to protect
+											 these files. */
+	
+	ends_in_ignored_extension(absolute_path, cfg)                           ||      /* filename ends in an extension
 											 * we were told to ignore */
-	found_under_dir(absolute_path, cfg->removable_media_mount_points) )             /* file is on a removable medium */
+	
+	(strlen(cfg->ignore_re) > 0 && matches_re(absolute_path, cfg->ignore_re)) ||   /* file name matches the IGNORE_RE */
+	
+	found_under_dir(absolute_path, cfg->removable_media_mount_points)       ||      /* file is on a removable medium */
+
+	is_empty_file(absolute_path, cfg)                                       )       /* zero byte-count in regular file */
 
      return BE_REMOVED;
 
@@ -1740,3 +1820,169 @@ static int move(const char *old_path, const char *new_path, config *cfg)
      return error1;
 }
 
+char * convert_relative_into_absolute_paths(const char *relative_paths)
+{
+   struct passwd *userinfo = NULL;
+
+   char *new_list = NULL;
+
+   const char *orig_ptr = NULL;
+   char *new_ptr = NULL;
+
+   unsigned int semicolon_count = 0;
+
+   userinfo = getpwuid(geteuid());
+
+   /* If it failed, we just return  NULL: */
+
+   if (!userinfo)
+     {
+#ifdef DEBUG
+	fprintf(stderr, "Unable to determine information about the user.\n");
+#endif
+	return NULL;
+     }
+
+   /* To determine the size of the buffer we must allocate, we must know how many semicolons there are in our argument: */
+
+   orig_ptr = relative_paths;
+
+   while (*orig_ptr != '\0')
+     {
+	if (*orig_ptr == ';')
+	  semicolon_count++;
+	orig_ptr++;
+     }
+
+   /* We allocate space for (i) the entire string we were passed, (ii) a trailing null char and (iii) a string holding
+    * the user's home dir and a slash for each relative path in our argument (the number of paths is given by semicolon_count+1): */
+
+   new_list = malloc(strlen(relative_paths) + 1 + (semicolon_count + 1) * (strlen(userinfo->pw_dir) + 1));
+
+   if (!new_list)
+     {
+#ifdef DEBUG
+	fprintf(stderr, "Memory allocation error inside convert_relative_into_absolute_paths().\n");
+#endif
+	return NULL;
+     }
+
+   /* Now we iterate along our argument, and copy its contents into the buffer new_list, always prefixing each dir name with the user's
+    * home directory: */
+
+   orig_ptr = relative_paths;
+   new_ptr = new_list;
+
+   /* First dir name also gets a prefix: */
+
+   strcpy(new_ptr, userinfo->pw_dir);
+   new_ptr += strlen(userinfo->pw_dir);
+   *new_ptr++ = '/';
+
+   while (*orig_ptr != '\0')
+     {
+	if (*orig_ptr != ';')
+	  *new_ptr = *orig_ptr;
+	else //if (*orig_ptr == ';')
+	  {
+	     *new_ptr++ = ';';
+	     strcat(new_ptr, userinfo->pw_dir);
+	     new_ptr += strlen(userinfo->pw_dir);
+	     *new_ptr = '/';
+	  }
+
+	orig_ptr++;
+	new_ptr++;
+     }
+
+   *new_ptr = '\0';
+
+   return new_list;
+}
+
+/* This function is used by decide_action to determine whether a
+ file is empty or not. It returns 1 if the file at path (i) exists, (ii) is a regular file
+ (or a symlink to one) and (iii) is empty, 0 otherwise. A config structure is needed so that
+ we can use real_fopen(). */
+
+static int is_empty_file(const char *path, config *cfg)
+{
+   FILE *fp;
+   struct stat file_stat;
+   int retval;
+
+   if (cfg->real_fopen_unavailable) /* if an error occurred, assume file contains data (ie, isn't empty). */
+     return 0;
+
+   retval = stat(path, &file_stat);
+
+   if (retval == -1 || /* stat() failed */
+       (!S_ISREG(file_stat.st_mode) && !S_ISLNK(file_stat.st_mode)) ) /* is neither a regular file nor a symlink */
+     return 0;
+
+   fp = cfg->real_fopen(path, "r"); /* open file and try to read the first byte. */
+
+   if (fp == NULL)
+     return 0;
+
+   if (fgetc(fp) == EOF && // no error occurred and file is empty
+       feof(fp) &&
+       !ferror(fp))
+     retval = 1;
+   else
+     retval = 0;
+
+   fclose(fp);
+
+   return retval;
+}
+
+/* The following two user-contributed functions implement support for the IGNORE_RE 
+ feature. */
+
+#ifdef DEBUG
+static void regex_report_error (int errcode, regex_t *compiled)
+{
+   size_t  len    = regerror (errcode, compiled, NULL, 0);
+   char   *buffer = malloc (len + 1);
+
+   regerror (errcode, compiled, buffer, len);
+   fprintf (stderr, "%s\n", buffer);
+   free (buffer);
+}
+#endif
+
+static int matches_re (const char *absolute_path, const char *regexp)
+{
+   regmatch_t matches[1];
+   regex_t    compiled;
+   int        ret;
+   
+   ret = regcomp (&compiled, regexp, REG_EXTENDED);
+
+   if (ret)
+     {
+#ifdef DEBUG
+	regex_report_error (ret, &compiled);
+#endif
+	regfree (&compiled);
+	return 0;
+     }
+
+   ret = regexec (&compiled, absolute_path, 1, matches, 0);
+
+#ifdef DEBUG
+   if (ret && ret != REG_NOMATCH)
+     {
+	regex_report_error (ret, &compiled);
+     }
+   if (ret == 0)
+     {
+	fprintf (stderr, "file %s matches %s\n", absolute_path, regexp);
+     }
+#endif
+
+   regfree (&compiled);
+
+   return (ret == 0);
+}
