@@ -1,4 +1,4 @@
-/* Copyright 2001, 2002 Manuel Arriaga
+/* Copyright 2001, 2002, 2003 Manuel Arriaga
  *
  *
  * This program is free software; you can redistribute it and/or modify
@@ -27,35 +27,49 @@
 #include <unistd.h>
 #include <dlfcn.h>
 
-/* Wrapper and helper functions are defined in this file: */
+/* Wrapper and helper functions are defined in this header file: */
 
 #include "trash.h"
 
+/* These are the buffers which hold the compile-time configuration defaults: */
+
+static char default_ignore_re[] = IGNORE_RE;
+
 static char default_ignore_extensions[] = IGNORE_EXTENSIONS;
 
 static char default_unremovable_dirs[] = UNREMOVABLE_DIRS;
 
+/* uncover_dirs doesn't show up here because there's no sense in having exceptions to unremovable_dirs by default... :-) */
+
 static char default_temporary_dirs[] = TEMPORARY_DIRS;
 
+static char default_user_temporary_dirs[] = USER_TEMPORARY_DIRS;
+
 static char default_removable_media_mount_points[] = REMOVABLE_MEDIA_MOUNT_POINTS;
 
+static char default_exceptions[] = EXCEPTIONS;
+
 static char default_relative_trash_can[] = TRASH_CAN;
 
 static char default_relative_trash_system_root[] = TRASH_SYSTEM_ROOT;
 
 /* calling init() is the first thing every wrapper function does. It is used
- to gather necessary user information and preferences (if any), perform
- some tests and "turn off" libtrash if it finds any problems. This is done
- by setting the variable general_failure, the value of which the wrapped
- functions check before running.  This function used to be called _init()
- so that it was automatically invoked by the linker whenever the library
- was loaded, but that caused two different problems: the initialization of
- other libraries (e.g., libproc when running 'ps') can occur before
- libtrash is initialized, and the functions we override must already be
- available at that time; from the moment in which the library is
- initialized to the point when the wrapper functions actually run the
- information _init() collected had already become out of date (e.g.,
- processes which change UIDs, etc). */
+ to gather necessary user information and preferences (if any),
+ perform some tests and "turn off" libtrash if it finds any problems.
+ This is done by setting the variable general_failure, the value of
+ which the wrapped functions check before running.  This function
+ used to be called _init() so that it was automatically invoked by
+ the linker whenever the library was loaded, but that caused two
+ different problems: (i) the initialization of other libraries (e.g.,
+ libproc when running 'ps') can occur before libtrash is initialized,
+ and the functions we override must already be available at that
+ time; (ii) from the moment in which the library is initialized to
+ the point when the wrapper functions actually run the information
+ _init() collected had already become out of date (e.g., processes
+ which change UIDs (<-Samba), etc). */
+
+/* init() is passed a pointer to a cfg structure (defined in trash.h), which it
+ * uses to communicate to the wrapper which called it the current configuration settings: */
 
 void init(config *cfg)
 {
@@ -68,6 +82,12 @@ void init(config *cfg)
 
    cfg->in_case_of_failure = IN_CASE_OF_FAILURE;
 
+   /* Holds a regular expression which causes files matching this r.e. to be
+    * ignored by libtrash:
+    */
+
+   cfg->ignore_re = default_ignore_re;
+
    /* Holds a list of file name extensions which cause files of these types to be ignored
     * by libtrash:
     */
@@ -116,11 +136,21 @@ void init(config *cfg)
 
    cfg->unremovable_dirs = default_unremovable_dirs;
 
+   /* This points to a list of exceptions from the above list of unremovable dirs (set to NULL at this point,
+    * because it doesn't have a default value -- it is meant to be used only at run-time): */
+
+   cfg->uncovered_dirs = NULL;
+
    /* This points to a semi-colon delimited list of directories under which the "real" functions
     * are always used, because they hold temporary files: */
 
    cfg->temporary_dirs = default_temporary_dirs;
 
+   /* This points to a semi-colon delimited list of directories *relative* to the user's home dir, under which
+    * the "real" functions are always used, because they hold temporary files: */
+
+   cfg->user_temporary_dirs = default_user_temporary_dirs;
+
    /* This points to a semi-colon delimited list of directories under which the "real" functions are
     * always used, because files under them are stored on removable media,
     * something which leads you to consider it not so worthwhile to keep a
@@ -128,6 +158,13 @@ void init(config *cfg)
 
    cfg->removable_media_mount_points = default_removable_media_mount_points;
 
+   /* This points to a list of paths (separated by semi-colons) which lists exceptions
+    * to libtrash protection: files which otherwise be considered either worthy of being saved
+    * in the trash can or unremovable lose their protection if listed here. This also applies to
+    * any files with a pathname which *starts* with a path listed in this list: */
+
+   cfg->exceptions = default_exceptions;
+
    /* 3- Paths: */
 
    /* Points to the absolute path of the directory in the user's home dir to which "deleted" files are moved: */
@@ -161,15 +198,15 @@ void init(config *cfg)
 
    /* Switches which individually turn on/off each of the wrapped functions: */
 
-   cfg->unlink_off = UNLINK_OFF;
+   cfg->intercept_unlink = INTERCEPT_UNLINK;
 
-   cfg->rename_off = RENAME_OFF;
+   cfg->intercept_rename = INTERCEPT_RENAME;
 
-   cfg->fopen_off  = FOPEN_OFF;
+   cfg->intercept_fopen  = INTERCEPT_FOPEN;
 
-   cfg->freopen_off = FREOPEN_OFF;
+   cfg->intercept_freopen = INTERCEPT_FREOPEN;
 
-   cfg->open_off   = OPEN_OFF;
+   cfg->intercept_open   = INTERCEPT_OPEN;
 
    /* Variables which "remember" if a pointer to a given GNU libc function is available: */
 
@@ -265,12 +302,59 @@ void init(config *cfg)
 	return;
      }
 
+   /* Has the user set UNCOVER_DIRS in her environment? */
+
+   tmp = getenv("UNCOVER_DIRS");
+
+   if (tmp)
+     {
+#ifdef DEBUG
+	fprintf(stderr, "UNCOVER_DIRS is set, directories %s won't be considered \"unremovable\".\n", tmp);
+#endif
+	cfg->uncovered_dirs = malloc(strlen(tmp) + 1);
+
+	if (cfg->uncovered_dirs) /* if malloc() failed, there's nothing we can do so
+				  * we just leave the pointer set to NULL -- that will
+				  * be interpreted as meaning that there are no uncovered dirs */
+	  {
+	     strcpy(cfg->uncovered_dirs, tmp);
+	  }
+     }
+
    /* ------------------------------------------- */
 
    /* Override compile-time defaults with values from the user-specific configuration file: */
 
    get_config_from_file(cfg);
 
+   /* Translate the relative paths in cfg->user_temporary_dirs into absolute paths: */
+
+   if (strlen(cfg->user_temporary_dirs) > 0)
+     {
+	tmp = convert_relative_into_absolute_paths(cfg->user_temporary_dirs);
+
+	if (tmp)
+	  cfg->user_temporary_dirs = tmp;
+	else  //  if convert_relative_... fails we signal failure
+	  {
+#ifdef DEBUG
+	     fprintf(stderr, "convert_relative_into_absolute_paths failed.\n");
+#endif
+
+	/* Free memory allocated inside get_config_from_files() which should be free()d inside this function: */
+
+	     if (cfg->relative_trash_can != default_relative_trash_can)
+	       free(cfg->relative_trash_can);
+
+	     if (cfg->relative_trash_system_root != default_relative_trash_system_root)
+	       free(cfg->relative_trash_system_root);
+
+	     cfg->general_failure = YES;
+
+	     return;
+	  }
+     }
+
    /* (All the configuration variables are by now set. The allocated memory which needs to be free()d will
     be deallocated in fini(), except for relative_trash_(can|system_root), which will be free()d before
     this function (init()) returns.) */
@@ -287,20 +371,27 @@ void init(config *cfg)
 	   "GLOBAL_PROTECTION:                 %d\n"
 	   "TRASH_SYSTEM_ROOT:                 %s\n"
 	   "TEMPORARY_DIRS:                    %s\n"
+	   "USER_TEMPORARY_DIRS:               %s\n"
 	   "UNREMOVABLE_DIRS:                  %s\n"
 	   "IGNORE_EXTENSIONS:                 %s\n"
-	   "UNLINK_OFF:                        %d\n"
-	   "RENAME_OFF:                        %d\n"
-	   "FOPEN_OFF:                         %d\n"
-	   "FREOPEN_OFF:                       %d\n"
-	   "OPEN_OFF:                          %d\n"
-	   "LIBTRASH_CONFIG_FILE_UNREMOVABLE: %d\n"
-	   "REMOVABLE_MEDIA_MOUNT_POINTS:      %s.\n\n",
+	   "INTERCEPT_UNLINK:                  %d\n"
+	   "INTERCEPT_RENAME:                  %d\n"
+	   "INTERCEPT_FOPEN:                   %d\n"
+	   "INTERCEPT_FREOPEN:                 %d\n"
+	   "INTERCEPT_OPEN:                    %d\n"
+	   "LIBTRASH_CONFIG_FILE_UNREMOVABLE:  %d\n"
+	   "REMOVABLE_MEDIA_MOUNT_POINTS:      %s\n"
+	   "EXCEPTIONS:                        %s\n"
+	   "IGNORE_RE:                         %s\n\n",
 	   cfg->relative_trash_can, cfg->in_case_of_failure, cfg->should_warn, cfg->ignore_hidden,
 	   cfg->ignore_editor_backup, cfg->ignore_editor_temporary, cfg->protect_trash, cfg->global_protection,
-	   cfg->relative_trash_system_root, cfg->temporary_dirs, cfg->unremovable_dirs, cfg->ignore_extensions,
-	   cfg->unlink_off, cfg->rename_off, cfg->fopen_off, cfg->freopen_off, cfg->open_off,
-	   cfg->libtrash_config_file_unremovable, cfg->removable_media_mount_points);
+	   cfg->relative_trash_system_root, cfg->temporary_dirs, cfg->user_temporary_dirs, cfg->unremovable_dirs,
+	   cfg->ignore_extensions,
+	   cfg->intercept_unlink, cfg->intercept_rename, cfg->intercept_fopen, cfg->intercept_freopen, cfg->intercept_open,
+	   cfg->libtrash_config_file_unremovable,
+	   cfg->removable_media_mount_points,
+	   cfg->exceptions,
+	   cfg->ignore_re);
 #endif
 
    /* ------------------------------------------------------ */
@@ -417,8 +508,8 @@ void init(config *cfg)
    if (!dir_ok(cfg->absolute_trash_can, NULL))
      {
 #ifdef DEBUG
-	fprintf(stderr, "The TRASH_CAN dir (%s) either doesn't exist (and its creation failed) or has insufficient
-			  permissions which we were unable to change.\ngeneral_failure set.\n", cfg->absolute_trash_can);
+	fprintf(stderr, "The TRASH_CAN dir (%s) either doesn't exist (and its creation failed) or has insufficient "
+		"permissions which we were unable to change.\ngeneral_failure set.\n", cfg->absolute_trash_can);
 #endif
 	cfg->general_failure = YES;
 	return;
@@ -427,8 +518,8 @@ void init(config *cfg)
    if (cfg->global_protection && !dir_ok(cfg->absolute_trash_system_root, NULL))
      {
 #ifdef DEBUG
-	fprintf(stderr, "The TRASH_CAN/SYSTEM_ROOT dir (%s) either doesn't exist (and its creation failed) or has insufficient
-			  permissions which we were unable to change.\ngeneral_failure set.\n", cfg->absolute_trash_system_root);
+	fprintf(stderr, "The TRASH_CAN/SYSTEM_ROOT dir (%s) either doesn't exist (and its creation failed) or has insufficient "
+		"permissions which we were unable to change.\ngeneral_failure set.\n", cfg->absolute_trash_system_root);
 #endif
 	cfg->general_failure = YES;
 	return;
@@ -441,7 +532,8 @@ void init(config *cfg)
 
 /* -------------------------------  */
 
-/* Just like init(), this function is always invoked by the wrapper functions before quitting: */
+/* Just like init(), this function is always invoked by the wrapper functions before quitting. Its
+ only task is freeing malloc()ed memory: */
 
 void fini(config *cfg)
 {
@@ -465,15 +557,27 @@ void fini(config *cfg)
    if (cfg->temporary_dirs != default_temporary_dirs)
      free(cfg->temporary_dirs);
 
+   if (cfg->user_temporary_dirs != default_user_temporary_dirs)
+     free(cfg->user_temporary_dirs);
+
    if (cfg->unremovable_dirs != default_unremovable_dirs)
      free(cfg->unremovable_dirs);
 
+   if (cfg->uncovered_dirs != NULL)
+     free(cfg->uncovered_dirs);
+
    if (cfg->removable_media_mount_points != default_removable_media_mount_points)
      free(cfg->removable_media_mount_points);
 
+   if (cfg->exceptions != default_exceptions)
+     free(cfg->exceptions);
+
    if (cfg->ignore_extensions != default_ignore_extensions)
      free(cfg->ignore_extensions);
 
+   if (cfg->ignore_re != default_ignore_re)
+     free(cfg->ignore_re);
+
    /* DON'T free() relative_trash_(can|system_root)! The memory they point to has already been free()d by
     * init() before returning. */
 
