@@ -1,4 +1,4 @@
-/* Copyright 2001, 2002 Manuel Arriaga
+/* Copyright 2001, 2002, 2003 Manuel Arriaga
  *
  *
  * This program is free software; you can redistribute it and/or modify
@@ -85,9 +85,12 @@
 
 #include "trash.h"
 
-#define FOPEN   1
-#define FREOPEN 2
-#define OPEN    3
+#define FOPEN     1
+#define FOPEN64   2
+#define FREOPEN   3
+#define FREOPEN64 4
+#define OPEN      5
+#define OPEN64    6
 
 /* This is the return type of the "magic" function, do_fopen_or_freopen_or_open(): */
 
@@ -104,7 +107,7 @@ static FdOrFp do_fopen_or_freopen_or_open(int function, const char *path, ...);
 
 /* -------------------- */
 
-/* These are the definitions of the wrappers for the three glibc functions we override: */
+/* These are the definitions of the wrappers for the six glibc functions we override: */
 
 FILE* fopen(const char *path, const char *mode)
 {
@@ -113,6 +116,13 @@ FILE* fopen(const char *path, const char *mode)
    return retval.fp; /* returns a file pointer */
 }
 
+FILE* fopen64(const char *path, const char *mode)
+{
+   FdOrFp retval = do_fopen_or_freopen_or_open(FOPEN64, path, mode);
+
+   return retval.fp; /* returns a file pointer */
+}
+
 FILE* freopen(const char *path, const char *mode, FILE *stream)
 {
    FdOrFp retval = do_fopen_or_freopen_or_open(FREOPEN, path, mode, stream);
@@ -120,6 +130,13 @@ FILE* freopen(const char *path, const char *mode, FILE *stream)
    return retval.fp; /* returns a file pointer */
 }
 
+FILE* freopen64(const char *path, const char *mode, FILE *stream)
+{
+   FdOrFp retval = do_fopen_or_freopen_or_open(FREOPEN64, path, mode, stream);
+
+   return retval.fp; /* returns a file pointer */
+}
+
 int open(const char *path, int flags, ...)
 {
    FdOrFp retval;
@@ -145,6 +162,33 @@ int open(const char *path, int flags, ...)
 
    return retval.fd; /* return an integer (file descriptor) */
 }
+
+int open64(const char *path, int flags, ...)
+{
+   FdOrFp retval;
+
+   /* If (flags & O_CREAT), then a third argument should be available: */
+
+   if (flags & O_CREAT)
+     {
+	va_list arg_list;
+
+	mode_t mode;
+
+	va_start(arg_list, flags);
+
+	mode = va_arg(arg_list, mode_t);
+
+	va_end(arg_list);
+
+	retval = do_fopen_or_freopen_or_open(OPEN64, path, flags, mode);
+     }
+   else /* we don't need to get a third argument: */
+     retval = do_fopen_or_freopen_or_open(OPEN64, path, flags);
+
+   return retval.fd; /* return an integer (file descriptor) */
+}
+
 /* ---------------------------------------------------------------------------- */
 
 /* Macros: */
@@ -158,10 +202,16 @@ int open(const char *path, int flags, ...)
 #define DETERMINE_FUNCTION_NAME()                                                         \
    if (function == FOPEN)                                                                 \
      function_name = fopen_name;                                                          \
+   else if (function == FOPEN64)                                                          \
+     function_name = fopen64_name;                                                        \
    else if (function == FREOPEN)                                                          \
      function_name = freopen_name;                                                        \
-   else /* if (function == OPEN) */                                                       \
+   else if (function == FREOPEN64)                                                        \
+     function_name = freopen64_name;                                                      \
+   else if (function == OPEN)                                                             \
      function_name = open_name;                                                           \
+   else if (function == OPEN64)                                                           \
+     function_name = open64_name;                                                         \
 
 /* -------------------------------------------------------------------- */
 
@@ -171,12 +221,13 @@ int open(const char *path, int flags, ...)
    {                                                                 \
       va_start(arg_list, path);                                      \
                                                                      \
-      if (function == FOPEN || function == FREOPEN)                  \
+      if (function == FOPEN || function == FREOPEN ||                \
+	     function == FOPEN64 || function == FREOPEN64)           \
 	mode_str = va_arg(arg_list, char*);                          \
-      if (function == FREOPEN)                                       \
+      if (function == FREOPEN || function == FREOPEN64)              \
 	stream = va_arg(arg_list, FILE*);                            \
       \
-      if (function == OPEN)                                          \
+      if (function == OPEN || function == OPEN64)                    \
 	{                                                            \
 	   flags = va_arg(arg_list, int);                            \
 	   if (flags & O_CREAT)                                      \
@@ -192,14 +243,15 @@ int open(const char *path, int flags, ...)
  to the value err (in most cases, it will be zero in order to avoid confusing the caller with meaningless errno values
  set by other library functions used by this meta-wrapper): */
 
-#define RETURN_FUNCTION_ERROR(err)                           \
-     {                                                       \
-	FdOrFp retval;                                       \
-	if (function == FOPEN || function == FREOPEN)        \
-	  retval.fp = NULL;                                  \
-	else /* if (function == OPEN) */                     \
-	  retval.fd = -1;                                    \
-	errno = err;                                         \
+#define RETURN_FUNCTION_ERROR(err)                              \
+     {                                                          \
+	FdOrFp retval;                                          \
+	if (function == FOPEN || function == FREOPEN ||         \
+	       function == FOPEN64 || function == FREOPEN64)    \
+	  retval.fp = NULL;                                     \
+	else /* if (function == OPEN || function == OPEN64)*/   \
+	  retval.fd = -1;                                       \
+	errno = err;                                            \
 	return retval; /* the caller will know which member of the union to access */ \
      }
 
@@ -207,12 +259,18 @@ int open(const char *path, int flags, ...)
 
 /* This is an ugly macro which invokes the real function with the right arguments and
  * evaluates to its return value: (Note: "real function" refers to the
- * function wrapped by the libindex function which called us.) */
+ * function wrapped by the libtrash function which called us.) */
 
-#define REAL_FUNCTION()                                                        \
-   (function == FOPEN ? (*real_fopen) (path, mode_str) :                       \
-       (function == FREOPEN ? (*real_freopen) (path, mode_str, stream) :       \
-	   ((flags & O_CREAT) ? (*real_open) (path, flags, mode) : (*real_open) (path, flags)) ) )
+#define REAL_FUNCTION()                                                             \
+   (function == FOPEN ? (*real_fopen) (path, mode_str) :                            \
+       (function == FOPEN64 ? (*real_fopen64) (path, mode_str) :                    \
+	   (function == FREOPEN ? (*real_freopen) (path, mode_str, stream) :        \
+	       (function == FREOPEN64 ? (*real_freopen64) (path, mode_str, stream) :\
+		   (function == OPEN ?                                              \
+		       (flags & O_CREAT ? (*real_open) (path, flags, mode) :        \
+			   (*real_open) (path, flags) ) :                           \
+		       (flags & O_CREAT ? (*real_open64) (path, flags, mode) :      \
+			   (*real_open64) (path, flags) ) ) ) ) ) )
 
 /* ------------------------------------------------------------------------------------- */
 
@@ -222,21 +280,23 @@ int open(const char *path, int flags, ...)
 #define RETURN_REAL_FUNCTION()                                  \
      {                                                          \
 	FdOrFp retval;                                          \
-	if (function == FOPEN || function == FREOPEN)           \
+	if (function == FOPEN || function == FREOPEN ||         \
+	       function == FOPEN64 || function == FREOPEN64)    \
 	  retval.fp = REAL_FUNCTION();                          \
-	else /* if (function == OPEN) */                        \
+	else /* if (function == OPEN || function == OPEN64) */  \
 	  retval.fd = REAL_FUNCTION();                          \
 	return retval;                                          \
      }
 
 /* --------------------------------------------------------------------------------------- */
 
-/* This macro evaluates to the *_off variable matching the function we are running
+/* This macro evaluates to the intercept_* variable matching the function we are running
  as: */
 
-#define REAL_FUNCTION_OFF()                                    \
-   (function == FOPEN ? cfg.fopen_off :                        \
-       (function == FREOPEN ? cfg.freopen_off : cfg.open_off) )
+#define REAL_FUNCTION_OFF()                                                       \
+   ((function == FOPEN || function == FOPEN64) ? !cfg.intercept_fopen :           \
+       ((function == FREOPEN || function == FREOPEN64) ? !cfg.intercept_freopen : \
+	   !cfg.intercept_open) )
 
 /* ------------------------------------------------------------------------------------------ */
 
@@ -258,8 +318,10 @@ int open(const char *path, int flags, ...)
  * truncating the specified file, to zero otherwise: */
 
 #define WRITE_TRUNCATE_MODE_ENABLED()                                                      \
-   ( ( (function == FOPEN || function == FREOPEN) && mode_str[0] == 'w') ||                \
-	(function == OPEN && (flags & O_WRONLY || flags & O_RDWR) && (flags & O_TRUNC) ) )
+   ( ( (function == FOPEN || function == FREOPEN ||                                        \
+	   function == FOPEN64 || function == FREOPEN64) && mode_str[0] == 'w' ) ||        \
+	( (function == OPEN || function == OPEN64) &&                                      \
+	    (flags & O_WRONLY || flags & O_RDWR) && (flags & O_TRUNC) ) )
 
 /* ----------------------------------------------------------------------------------------------- */
 
@@ -277,15 +339,15 @@ int open(const char *path, int flags, ...)
 
 #define TEST_FILE_EXISTENCE()                                                                           \
      {                                                                                                  \
-	if ( (function == OPEN) && (flags & O_NOFOLLOW) )                                               \
+	if ( (function == OPEN || function == OPEN64) && (flags & O_NOFOLLOW) )                                               \
 	  error = lstat(path, &file_stat);                                                              \
 	else                                                                                            \
 	  error = stat(path, &file_stat);                                                               \
-	                                                                                                \
+	\
 	if ( (error && errno == ENOENT)                 ||                                              \
 		(!error && S_ISDIR(file_stat.st_mode))  ||                                              \
 		(!error && !S_ISREG(file_stat.st_mode) && !S_ISLNK(file_stat.st_mode)) ||               \
-		(!error && (function == OPEN) && (flags & O_NOFOLLOW) && S_ISLNK(file_stat.st_mode) ) ) \
+		(!error && (function == OPEN || function == OPEN64) && (flags & O_NOFOLLOW) && S_ISLNK(file_stat.st_mode) ) ) \
 	  {                                                                                             \
 	     fini(&cfg);                                                                                \
 	     RETURN_REAL_FUNCTION();                                                                    \
@@ -301,13 +363,12 @@ int open(const char *path, int flags, ...)
  * will need to create a new file if it is going to succeed. For that reason, if the user didn't originally set
  * O_CREAT in flags, we need to do it ourselves so that the call to the real open() results in the (re)creation of
  * that file and that function succeeds. But, when instructing open() to create a new file if necessary, we also need
- * to tell it which mode to use when creating it. Again, if the user didn't specify one, we will need to do it
- * ourselves - and which mode should we choose? Well, we will instruct open() to create a new file with the same
+ * to tell it which mode to use when creating it. We will instruct open() to create a new file with the same
  * permissions (mode) which the original one had: */
 
 #define DO_OPEN_SPECIFIC_PREPARATION()                                       \
    {                                                                         \
-      if (function == OPEN && !(flags & O_CREAT))                            \
+      if (function == OPEN || function == OPEN64)                            \
 	{                                                                    \
 	   flags |= O_CREAT;                                                 \
 	   mode = file_stat.st_mode;                                         \
@@ -328,17 +389,28 @@ int open(const char *path, int flags, ...)
  * to the rest of the function) */
 
 #define GET_REAL_FUNCTION()                                                            \
-      int (*real_fopen) (const char *path, const char *mode) = NULL;                   \
-      int (*real_freopen) (const char *path, const char *mode, FILE *stream) = NULL;   \
-      int (*real_open) (const char *path, int flags, ...) = NULL;                      \
-      if (function == FOPEN)                                                           \
-	real_fopen = dlsym(RTLD_NEXT, "fopen");                                        \
-      else if (function == FREOPEN)                                                    \
-	real_freopen = dlsym(RTLD_NEXT, "freopen");                                    \
-      else /* if (function == OPEN) */                                                 \
-	real_open = dlsym(RTLD_NEXT, "open");                                          \
-      if (dlerror())                                                                   \
-	RETURN_FUNCTION_ERROR(0);                                                      \
+   int (*real_fopen) (const char *path, const char *mode) = NULL;                      \
+   int (*real_fopen64) (const char *path, const char *mode) = NULL;                    \
+   int (*real_freopen) (const char *path, const char *mode, FILE *stream) = NULL;      \
+   int (*real_freopen64) (const char *path, const char *mode, FILE *stream) = NULL;    \
+   int (*real_open) (const char *path, int flags, ...) = NULL;                         \
+   int (*real_open64) (const char *path, int flags, ...) = NULL;                       \
+   \
+   if (function == FOPEN)                                                           \
+     real_fopen = dlsym(RTLD_NEXT, "fopen");                                        \
+   else if (function == FOPEN64)                                                    \
+     real_fopen64 = dlsym(RTLD_NEXT, "fopen64");                                    \
+   else if (function == FREOPEN)                                                    \
+     real_freopen = dlsym(RTLD_NEXT, "freopen");                                    \
+   else if (function == FREOPEN64)                                                  \
+     real_freopen64 = dlsym(RTLD_NEXT, "freopen64");                                \
+   else if (function == OPEN)                                                       \
+     real_open = dlsym(RTLD_NEXT, "open");                                          \
+   else /* if (function == OPEN64) */                                               \
+     real_open64 = dlsym(RTLD_NEXT, "open64");                                      \
+   \
+   if (dlerror())                                                                   \
+     RETURN_FUNCTION_ERROR(0);                                                      \
 
 /* ----------------------------------------------------------------------------- */
 
@@ -369,9 +441,14 @@ static FdOrFp do_fopen_or_freopen_or_open(int function, const char *path, ...)
 
    int file_should = 0;
 
-   const char *fopen_name = "fopen", *freopen_name = "freopen", *open_name = "open";
+#ifdef DEBUG
+   const char
+     *fopen_name   = "fopen",   *fopen64_name = "fopen64",
+     *freopen_name = "freopen", *freopen64_name = "freopen64",
+     *open_name    = "open",    *open64_name = "open64";
 
    const char *function_name = NULL;
+#endif
 
    config cfg;
 
@@ -422,12 +499,12 @@ static FdOrFp do_fopen_or_freopen_or_open(int function, const char *path, ...)
 
    /* From this point on, we always call fini() before quittting. */
 
-   /* Is cfg.libtrash_off or cfg.(real_function)_off set? If so, just invoke the real function: */
+   /* Is cfg.libtrash_off set to true or cfg.intercept_(real_function) set to false? If so, just invoke the real function: */
 
    if (cfg.libtrash_off || REAL_FUNCTION_OFF())
      {
 #ifdef DEBUG
-	fprintf(stderr, "Passing request to the real function because (libtrash|%s)_off is set.\n", function_name);
+	fprintf(stderr, "Passing request to the real function because libtrash_off = true or intercept_%s = false.\n", function_name);
 #endif
 	fini(&cfg);
 	RETURN_REAL_FUNCTION();
